/* $Id: cpl_image_filter-test.c,v 1.23 2008/02/07 10:54:01 llundin Exp $
 *
 * This file is part of the ESO Common Pipeline Library
 * Copyright (C) 2001-2004 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2008/02/07 10:54:01 $
 * $Revision: 1.23 $
 * $Name:  $
 */

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <math.h>
#include <float.h>
#include <assert.h>

#include "cpl_image_filter.h"
#include "cpl_image_gen.h"
#include "cpl_image_io.h"
#include "cpl_tools.h"
#include "cpl_image_bpm.h"
#include "cpl_memory.h"

/*-----------------------------------------------------------------------------
                                Defines
 -----------------------------------------------------------------------------*/

#define CPL_IMAGE_FILTER_LINEAR 0
#define CPL_IMAGE_FILTER_MORPHO 1
#define CPL_IMAGE_FILTER_MEDIAN 2
#define CPL_IMAGE_FILTER_STDEV  3
#define CPL_IMAGE_FILTER_SIZE   4

#ifndef IMAGE_SIZE_X
#define IMAGE_SIZE_X 256
#endif
#ifndef IMAGE_SIZE_Y
#define IMAGE_SIZE_Y IMAGE_SIZE_X
#endif

/*-----------------------------------------------------------------------------
                        Private function prototypes
 -----------------------------------------------------------------------------*/

static void filter_one(const cpl_image *, int, int, int, double *, cpl_flops *);

static const char * filter_txt[CPL_IMAGE_FILTER_SIZE] =
    {"linear", "morpho", "median", "stdev"};

static const int filter_max[CPL_IMAGE_FILTER_SIZE] = { 31, 5, 9, 31};
static const cpl_boolean has_nm[CPL_IMAGE_FILTER_SIZE] =
    { CPL_TRUE, CPL_FALSE, CPL_TRUE, CPL_TRUE };

/*-----------------------------------------------------------------------------
                                  Main
 -----------------------------------------------------------------------------*/
int main(void)
{
    int itype;
    cpl_type    pixeltype = CPL_TYPE_FLOAT;
    double      cputime[CPL_IMAGE_FILTER_SIZE] = {0.0, 0.0, 0.0, 0.0};
    cpl_flops   nflops[CPL_IMAGE_FILTER_SIZE] = {0, 0, 0, 0};
    const int   nbad = 10;
    /* the bad pixels positions */
    const int   bad_pos_x[] = {4, 5, 6, 4, 5, 6, 4, 5, 6, 8};
    const int   bad_pos_y[] = {5, 5, 5, 6, 6, 6, 7, 7, 7, 9};
    const int   nx = IMAGE_SIZE_X;
    const int   ny = IMAGE_SIZE_Y;


    cpl_test_init(PACKAGE_BUGREPORT, CPL_MSG_WARNING);

    /* Insert tests below */

    cpl_tools_get_cputime(CPL_CLOCK_STOP); /* STOP/START in filter_one() */

    for (itype = 0; itype < 2; itype++, pixeltype = CPL_TYPE_DOUBLE) {
        int do_bad;
        int filter_type;

        cpl_image * raw = cpl_image_new(nx, ny, pixeltype);

        cpl_msg_info(cpl_func, "Filtering %d x %d image of pixel type %s", nx,
                     ny, pixeltype == CPL_TYPE_DOUBLE ? "double" : "float");

        cpl_test_zero(cpl_image_fill_noise_uniform(raw, 0.0, 1.0));

        for (do_bad = 0; do_bad < 2; do_bad++) {

            if (do_bad) {
                int i;
                cpl_msg_info(cpl_func, "Rejecting %d pixels", nbad);
                for (i=0; i < nbad; i++)
                    cpl_test_zero(cpl_image_reject(raw, bad_pos_x[i], bad_pos_y[i]));
            }

            for (filter_type = CPL_IMAGE_FILTER_LINEAR;
                 filter_type < CPL_IMAGE_FILTER_SIZE; filter_type++) {

                /* TEST 1x1 kernel filtering */
                filter_one(raw, 1, 1, filter_type, &(cputime[filter_type]),
                           &(nflops[filter_type]));

                if (has_nm[filter_type]) {
                    /* TEST 3x1 kernel filtering */

                    filter_one(raw, 3, 1, filter_type, &(cputime[filter_type]),
                               &(nflops[filter_type]));

                    /* TEST 1x3 kernel filtering */
                    filter_one(raw, 1, 3, filter_type, &(cputime[filter_type]),
                               &(nflops[filter_type]));

                }

                /* TEST 3x3 kernel filtering */

                filter_one(raw, 3, 3, filter_type, &(cputime[filter_type]),
                           &(nflops[filter_type]));

                /* TEST max size kernel filtering */
                filter_one(raw, filter_max[filter_type],
                           filter_max[filter_type], filter_type,
                           &(cputime[filter_type]), &(nflops[filter_type]));

                if (nflops[filter_type] == 0 || cputime[filter_type] <= 0.0)
                    continue;
                cpl_msg_info(cpl_func, "%d x %d image %s filter total "
                             "[MFLOP/s]: %g", nx, ny, filter_txt[filter_type],
                             nflops[filter_type] / cputime[filter_type] * 1e-6);
            }
        }
        cpl_image_delete(raw);
    }

    cpl_tools_get_cputime(CPL_CLOCK_START);

    return cpl_test_end(0);
}



/*----------------------------------------------------------------------------*/
/**
  @brief  Filter an image with a M x N unity/uniform kernel
  @param  self    The image to filter
  @param  m       The kernel X-size
  @param  n       The kernel Y-size
  @param  type    The filtering type
  @param  ptime   Add the (CPU-)time spent to *ptime
  @param  pflops  Add the FLOPs done to *pflops
  @return void

 */
/*----------------------------------------------------------------------------*/
static void filter_one(const cpl_image * self, int m, int n, int type,
                       double * ptime, cpl_flops * pflops)
{

    cpl_image    * filtered;
    cpl_image    * (*filter_func) (const cpl_image *, const cpl_matrix *);
    cpl_matrix   * kernel = cpl_matrix_new(m, n);
    const cpl_type pixel = cpl_image_get_type(self);
    cpl_flops      flops;
    double         cputime;


    switch (type) {
    case CPL_IMAGE_FILTER_LINEAR:
        {
            filter_func = cpl_image_filter_linear;
            break;
        }
    case CPL_IMAGE_FILTER_MORPHO:
        {
            filter_func = cpl_image_filter_morpho;
            break;
        }
    case CPL_IMAGE_FILTER_MEDIAN:
        {
            filter_func = cpl_image_filter_median;
            break;
        }
    case CPL_IMAGE_FILTER_STDEV:
        {
            filter_func = cpl_image_filter_stdev;
            break;
        }
    default:
        {
            assert(0);
            filter_func = NULL;
        }
    }

    cpl_test_zero( (type == CPL_IMAGE_FILTER_LINEAR
                ? cpl_matrix_set(kernel, m/2, n/2, 1.0)
                : cpl_matrix_fill(kernel, 1.0)));

    flops = cpl_tools_get_flops();
    cpl_tools_get_cputime(CPL_CLOCK_START);

    filtered = filter_func(self, kernel);

    cputime = cpl_tools_get_cputime(CPL_CLOCK_STOP);

    flops = cpl_tools_get_flops() - flops;

    cpl_matrix_delete(kernel);

    cpl_test_nonnull(filtered);
    if (type == CPL_IMAGE_FILTER_LINEAR) {
        double sqflux;
        cpl_test_zero(cpl_image_subtract(filtered, self));
        sqflux = cpl_image_get_sqflux(filtered)/DBL_EPSILON;
        cpl_msg_info(cpl_func, "%s image linear %d x %d filtering error "
                     "[DBL_EPSILON]: %g",
                     pixel == CPL_TYPE_DOUBLE ? "double" : "float",
                     m, n, sqflux);
        cpl_test_leq(sqflux, sqrt(IMAGE_SIZE_X * IMAGE_SIZE_Y));
    }

    cpl_image_delete(filtered);

    *ptime += cputime;
    *pflops += flops;

    cpl_msg_info(cpl_func, "%s image %s %d x %d filtering [s]: %g",
                 pixel == CPL_TYPE_DOUBLE ? "double" : "float",
                 filter_txt[type], m, n, cputime);

    if (flops > 0 && cputime > 0.0)
        cpl_msg_info(cpl_func, "%s image %s %d x %d filtering [MFLOP/s]: %g",
                     pixel == CPL_TYPE_DOUBLE ? "double" : "float",
                     filter_txt[type], m, n, flops / cputime * 1e-6);

    return;
}
