/* $Id: cpl_imagelist_io-test.c,v 1.29 2008/02/07 10:54:01 llundin Exp $
 *
 * This file is part of the ESO Common Pipeline Library
 * Copyright (C) 2001-2004 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2008/02/07 10:54:01 $
 * $Revision: 1.29 $
 * $Name:  $
 */

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include "cpl_imagelist_io.h"
#include "cpl_image_gen.h"
#include "cpl_image_io.h"
#include "cpl_vector.h"
#include "cpl_tools.h"
#include "cpl_memory.h"

/*-----------------------------------------------------------------------------
                                   Define
 -----------------------------------------------------------------------------*/

#define     IMAGE_SZ    127
#define     NIMAGES     10

/*----------------------------------------------------------------------------*/
/**
  @brief    Empty an imagelist and deallocate all its images
  @param    self  The image list
 */
/*----------------------------------------------------------------------------*/
static void cpl_imagelist_empty(cpl_imagelist * self)
{

    if (self == NULL) {
        cpl_error_set(cpl_func, CPL_ERROR_NULL_INPUT);
    } else {
        int i = cpl_imagelist_get_size(self);
        while (i--) {
            cpl_image_delete(cpl_imagelist_unset(self, i));
        }
    }
    return;
}

/*-----------------------------------------------------------------------------
                                  Main
 -----------------------------------------------------------------------------*/
int main(void)
{
    const char    * file = "cpl_imagelist_test.fits";
    cpl_imagelist * imlist;
    cpl_imagelist * imlist2;
    cpl_image     * image;
    double          flags[NIMAGES];
    cpl_vector    * eraser;
    int             i;


    cpl_test_init(PACKAGE_BUGREPORT, CPL_MSG_WARNING);

    /* Insert tests below */

    eraser = cpl_vector_wrap(NIMAGES, flags);

    cpl_tools_get_cputime(CPL_CLOCK_START);

    cpl_msg_info("", "Checking various failures");

    image = cpl_image_fill_test_create(IMAGE_SZ, IMAGE_SZ);
    cpl_test_nonnull( image );

    /* Test cpl_imagelist_new() */
    imlist = cpl_imagelist_new();
    cpl_test_nonnull( imlist );

    cpl_test( cpl_imagelist_get_size(imlist) == 0 );
    cpl_test( cpl_imagelist_is_uniform(imlist) == 1);

    /* Failure tests involving NULL end empty lists */
    cpl_test( cpl_imagelist_get_size(NULL) );
    cpl_test_error( CPL_ERROR_NULL_INPUT );

    cpl_test( cpl_imagelist_is_uniform(NULL) < 0);
    cpl_test_error( CPL_ERROR_NULL_INPUT );

    cpl_test( cpl_imagelist_set(NULL, image, 0) == CPL_ERROR_NULL_INPUT );
    cpl_test_error( CPL_ERROR_NULL_INPUT );

    cpl_test( cpl_imagelist_set(imlist, NULL, 0) == CPL_ERROR_NULL_INPUT );
    cpl_test_error( CPL_ERROR_NULL_INPUT );

    cpl_test( cpl_imagelist_set(imlist, image, -1)
                       == CPL_ERROR_ILLEGAL_INPUT );
    cpl_test_error( CPL_ERROR_ILLEGAL_INPUT );

    cpl_test( cpl_imagelist_set(imlist, image, 1)
                       == CPL_ERROR_ACCESS_OUT_OF_RANGE );
    cpl_test_error( CPL_ERROR_ACCESS_OUT_OF_RANGE );

    cpl_test( cpl_imagelist_get(NULL, 0) == NULL );
    cpl_test_error( CPL_ERROR_NULL_INPUT );

    cpl_test( cpl_imagelist_get(imlist, -1) == NULL );
    cpl_test_error( CPL_ERROR_ILLEGAL_INPUT );

    cpl_test( cpl_imagelist_get(imlist, 0) == NULL );
    cpl_test_error( CPL_ERROR_ACCESS_OUT_OF_RANGE );

    cpl_test( cpl_imagelist_erase(imlist, NULL)
                       == CPL_ERROR_NULL_INPUT );
    cpl_test_error( CPL_ERROR_NULL_INPUT );

    /* The elements of eraser of not initialized at this point,
       but they are also not supposed to be accessed */
    cpl_test( cpl_imagelist_erase(NULL, eraser)
                       == CPL_ERROR_NULL_INPUT );
    cpl_test_error( CPL_ERROR_NULL_INPUT );

    cpl_test( cpl_imagelist_erase(imlist, eraser)
                       == CPL_ERROR_INCOMPATIBLE_INPUT );
    cpl_test_error( CPL_ERROR_INCOMPATIBLE_INPUT );

    cpl_test( cpl_imagelist_duplicate(NULL) == NULL );
    cpl_test_error( CPL_ERROR_NULL_INPUT );

    cpl_test( cpl_imagelist_save(imlist, NULL,  CPL_BPP_IEEE_FLOAT, NULL,
                                          CPL_IO_DEFAULT)
                       == CPL_ERROR_NULL_INPUT );
    cpl_test_error( CPL_ERROR_NULL_INPUT );

    cpl_test( cpl_imagelist_save(NULL,  file, CPL_BPP_IEEE_FLOAT, NULL,
                                          CPL_IO_DEFAULT)
                       == CPL_ERROR_NULL_INPUT );
    cpl_test_error( CPL_ERROR_NULL_INPUT );

    cpl_test( cpl_imagelist_save(imlist, file,  CPL_BPP_IEEE_FLOAT, NULL,
                                          CPL_IO_DEFAULT)
                       == CPL_ERROR_ILLEGAL_INPUT );
    cpl_test_error( CPL_ERROR_ILLEGAL_INPUT );

    cpl_test( cpl_imagelist_load(NULL, CPL_TYPE_INT, 0) == NULL );
    cpl_test_error( CPL_ERROR_NULL_INPUT );

    cpl_test( cpl_imagelist_load(file, CPL_TYPE_INT, -1) == NULL );
    cpl_test_error( CPL_ERROR_ILLEGAL_INPUT );

    cpl_test( cpl_imagelist_load(".", CPL_TYPE_INT, 0) == NULL );
    cpl_test_error( CPL_ERROR_FILE_IO );


    cpl_test( cpl_imagelist_load_window(NULL, CPL_TYPE_INT, 0, 1, 1,
                                                 2, 2) == NULL );
    cpl_test_error( CPL_ERROR_NULL_INPUT );

    cpl_test( cpl_imagelist_load_window(file, CPL_TYPE_INT, -1, 1, 1,
                                                 2, 2) == NULL );
    cpl_test_error( CPL_ERROR_ILLEGAL_INPUT );

    cpl_test( cpl_imagelist_load_window(".", CPL_TYPE_INT, 0, 1, 1,
                                                 2, 2) == NULL );
    cpl_test_error( CPL_ERROR_FILE_IO );


    /* Test cpl_imagelist_duplicate() of empty list */
    imlist2 = cpl_imagelist_duplicate(imlist);
    cpl_test_nonnull( imlist2 );
    
    cpl_test( cpl_imagelist_get_size(imlist2) == 0 );
    cpl_test( cpl_imagelist_is_uniform(imlist2) == 1);

    cpl_imagelist_empty(imlist2); /* Test cpl_imagelist_unset() */
    cpl_imagelist_delete(imlist2);
    
    cpl_msg_info("", "Create an image list of %d images", NIMAGES);

    /* Test cpl_imagelist_set() */
    for (i=0; i < NIMAGES ; i++) {
        cpl_image * copy = cpl_image_duplicate(image);

        flags[i] = i % 2 ? 1.0 : -1.0;

        cpl_test_zero( cpl_imagelist_set(imlist, copy, i) );

        cpl_test( cpl_imagelist_is_uniform(imlist) == 0 );
        cpl_test( cpl_imagelist_get_size(imlist) == i+1 );
        cpl_test( cpl_imagelist_get(imlist, i) == copy );
    
    }
    cpl_image_delete(image);

    cpl_msg_info("", "Duplicate the image list");

    imlist2 = cpl_imagelist_duplicate(imlist);
    cpl_test_nonnull( imlist2 );

    cpl_test( cpl_imagelist_is_uniform(imlist2) == 0);
    cpl_test( cpl_imagelist_get_size(imlist2) == NIMAGES );

    /* Failure tests involving non-empty, valid list */
    cpl_test( cpl_imagelist_save(imlist2, ".",  CPL_BPP_IEEE_FLOAT, NULL,
                                          CPL_IO_DEFAULT)
                       == CPL_ERROR_FILE_IO );
    cpl_test_error( CPL_ERROR_FILE_IO );

    image = cpl_image_new(IMAGE_SZ, IMAGE_SZ, CPL_TYPE_INT);
    cpl_test_nonnull( image );

    cpl_test( cpl_imagelist_set(imlist2, image, 0)
                       == CPL_ERROR_TYPE_MISMATCH );
    cpl_test_error( CPL_ERROR_TYPE_MISMATCH );
    cpl_test( cpl_imagelist_is_uniform(imlist2) == 0 );
    cpl_test( cpl_imagelist_get_size(imlist2) == NIMAGES );

    cpl_image_delete(image);

    image = cpl_image_new(IMAGE_SZ/2, IMAGE_SZ, CPL_TYPE_INT);
    cpl_test_nonnull( image );

    cpl_test( cpl_imagelist_set(imlist2, image, 0)
                       == CPL_ERROR_INCOMPATIBLE_INPUT );
    cpl_test_error( CPL_ERROR_INCOMPATIBLE_INPUT );
    cpl_test( cpl_imagelist_is_uniform(imlist2) == 0 );
    cpl_test( cpl_imagelist_get_size(imlist2) == NIMAGES );

    cpl_image_delete(image);

    cpl_test( cpl_imagelist_get(imlist2, NIMAGES+1) == NULL);
    cpl_test_error( CPL_ERROR_ACCESS_OUT_OF_RANGE );

    cpl_test( cpl_imagelist_is_uniform(imlist2) == 0 );
    cpl_test( cpl_imagelist_get_size(imlist2) == NIMAGES );

    /* Normal conditions */

    /* Test cpl_imagelist_save() */
    cpl_msg_info("", "Save the image list to %s", file);
    cpl_test_zero( cpl_imagelist_save(imlist, file,  CPL_BPP_IEEE_FLOAT, NULL,
                                          CPL_IO_DEFAULT) );
    cpl_test_zero( cpl_imagelist_save(imlist, file,  CPL_BPP_IEEE_FLOAT, NULL,
                                          CPL_IO_EXTEND) );

    cpl_test_zero( cpl_imagelist_erase(imlist, eraser) );
    cpl_test( cpl_imagelist_is_uniform(imlist) == 0 );
    cpl_test( 2*cpl_imagelist_get_size(imlist) == NIMAGES );

    cpl_imagelist_empty(imlist); /* Test cpl_imagelist_unset() */
    cpl_imagelist_delete(imlist);
    
    /* Test cpl_imagelist_load() */
    cpl_msg_info("", "Load image list as type DOUBLE");
    imlist = cpl_imagelist_load(file, CPL_TYPE_DOUBLE, 0);

    cpl_test_nonnull( imlist );
    cpl_test( cpl_imagelist_is_uniform(imlist) == 0 );
    cpl_test( cpl_imagelist_get_size(imlist) == NIMAGES );

    cpl_test_zero( cpl_imagelist_erase(imlist, eraser) );
    cpl_test( cpl_imagelist_is_uniform(imlist) == 0 );
    cpl_test( 2*cpl_imagelist_get_size(imlist) == NIMAGES );

    cpl_imagelist_empty(imlist); /* Test cpl_imagelist_unset() */
    cpl_imagelist_delete(imlist);
    
    /* Test cpl_imagelist_load() */
    cpl_msg_info("", "Load image list as type FLOAT");
    imlist = cpl_imagelist_load(file, CPL_TYPE_FLOAT, 0);

    cpl_test_nonnull( imlist );
    cpl_test( cpl_imagelist_is_uniform(imlist) == 0 );
    cpl_test( cpl_imagelist_get_size(imlist) == NIMAGES );

    cpl_test_zero( cpl_imagelist_erase(imlist, eraser) );
    cpl_test( cpl_imagelist_is_uniform(imlist) == 0 );
    cpl_test( 2*cpl_imagelist_get_size(imlist) == NIMAGES );

    cpl_imagelist_empty(imlist); /* Test cpl_imagelist_unset() */
    cpl_imagelist_delete(imlist);

    /* Test cpl_imagelist_load() */
    cpl_msg_info("", "Load image list as type INTEGER");
    imlist = cpl_imagelist_load(file, CPL_TYPE_INT, 0);

    cpl_test_nonnull( imlist );
    cpl_test( cpl_imagelist_is_uniform(imlist) == 0 );
    cpl_test( cpl_imagelist_get_size(imlist) == NIMAGES );

    cpl_test_zero( cpl_imagelist_erase(imlist, eraser) );
    cpl_test( cpl_imagelist_is_uniform(imlist) == 0 );
    cpl_test( 2*cpl_imagelist_get_size(imlist) == NIMAGES );

    cpl_imagelist_empty(imlist); /* Test cpl_imagelist_unset() */
    cpl_imagelist_delete(imlist);

    /* Test cpl_imagelist_load_window() */
    cpl_msg_info("", "Load image list as type DOUBLE from a window");
    imlist = cpl_imagelist_load_window(file, CPL_TYPE_DOUBLE, 0, IMAGE_SZ/4,
                                       IMAGE_SZ/4, 3*IMAGE_SZ/4, 3*IMAGE_SZ/4);

    cpl_test_nonnull( imlist );
    cpl_test( cpl_imagelist_is_uniform(imlist) == 0 );
    cpl_test( cpl_imagelist_get_size(imlist) == NIMAGES );

    cpl_test_zero( cpl_imagelist_erase(imlist, eraser) );
    cpl_test( cpl_imagelist_is_uniform(imlist) == 0 );
    cpl_test( 2*cpl_imagelist_get_size(imlist) == NIMAGES );

    cpl_imagelist_empty(imlist); /* Test cpl_imagelist_unset() */
    cpl_imagelist_delete(imlist);
    

    /* Test cpl_imagelist_load_window() */
    cpl_msg_info("", "Load image list as type FLOAT from a window");
    imlist = cpl_imagelist_load_window(file, CPL_TYPE_FLOAT, 0, IMAGE_SZ/4,
                                       IMAGE_SZ/4, 3*IMAGE_SZ/4, 3*IMAGE_SZ/4);

    cpl_test_nonnull( imlist );
    cpl_test( cpl_imagelist_is_uniform(imlist) == 0 );
    cpl_test( cpl_imagelist_get_size(imlist) == NIMAGES );

    cpl_test_zero( cpl_imagelist_erase(imlist, eraser) );
    cpl_test( cpl_imagelist_is_uniform(imlist) == 0 );
    cpl_test( 2*cpl_imagelist_get_size(imlist) == NIMAGES );

    cpl_imagelist_empty(imlist); /* Test cpl_imagelist_unset() */
    cpl_imagelist_delete(imlist);
    

    /* Test cpl_imagelist_load_window() */
    cpl_msg_info("", "Load image list as type INTEGER from a window");
    imlist = cpl_imagelist_load_window(file, CPL_TYPE_INT, 0, IMAGE_SZ/4,
                                       IMAGE_SZ/4, 3*IMAGE_SZ/4, 3*IMAGE_SZ/4);

    cpl_test_nonnull( imlist );
    cpl_test( cpl_imagelist_is_uniform(imlist) == 0 );
    cpl_test( cpl_imagelist_get_size(imlist) == NIMAGES );

    cpl_test_zero( cpl_imagelist_erase(imlist, eraser) );
    cpl_test( cpl_imagelist_is_uniform(imlist) == 0 );
    cpl_test( 2*cpl_imagelist_get_size(imlist) == NIMAGES );

    cpl_imagelist_empty(imlist); /* Test cpl_imagelist_unset() */
    cpl_imagelist_delete(imlist);

    /* Clean up  */

    cpl_imagelist_empty(imlist2); /* Test cpl_imagelist_unset() */
    cpl_imagelist_delete(imlist2);
    cpl_vector_unwrap(eraser);
    remove(file);

    /* End of tests */
    return cpl_test_end(0);
}
